from parsagon.executor import Executor


class MockDriver:
    def __init__(self, current_url, page_source):
        self.current_url = current_url
        self.page_source = page_source


class MockExecutor(Executor):
    def __init__(self, current_url, page_source):
        self.driver = MockDriver(current_url, page_source)
        self.max_elem_id = 0
        self.custom_functions = {}


def test_makes_links_absolute():
    executor = MockExecutor("https://example.com/", '<html><body><a href="/stuff">Stuff</a></body></html>')
    root = executor._get_cleaned_lxml_root()
    assert root[0][0].get("href") == "https://example.com/stuff"



from json import JSONDecodeError

import httpx

from parsagon import settings
from parsagon.exceptions import APIException, ProgramNotFoundException

environment = "PANDAS_1.x"


class RaiseProgramNotFound:
    def __init__(self, program_name):
        self.program_name = program_name

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None and issubclass(exc_type, APIException):
            if exc_value.status_code == 404:
                raise ProgramNotFoundException(self.program_name)
        return False


def _request_to_exception(response):
    status_code = response.status_code
    if status_code == 500:
        raise APIException("A server error occurred. Please notify Parsagon.", status_code)
    if status_code in (502, 503, 504):
        raise APIException("Lost connection to server.", status_code)
    try:
        errors = response.json()
        if "non_field_errors" in errors:
            raise APIException(errors["non_field_errors"], status_code)
        else:
            raise APIException(errors, status_code)
    except JSONDecodeError:
        raise APIException("Could not parse response.", status_code)


def _api_call(httpx_func, endpoint, **kwargs):
    api_key = settings.get_api_key()
    api_endpoint = f"{settings.get_api_base()}/api{endpoint}"
    headers = {"Authorization": f"Token {api_key}"}
    r = httpx_func(api_endpoint, headers=headers, timeout=None, **kwargs)
    if not r.is_success:
        _request_to_exception(r)
    else:
        try:
            return r.json()
        except:
            return None


def get_program_sketches(description):
    """
    Gets a program sketches (full and abridged) from a description.
    :param description: Description in natural language that will be used to generate the scraping program.
    :return: A dict with keys "full", "abridged", and "pseudocode" for the respective program ASTs and pseudocode.
    """
    return _api_call(httpx.post, "/transformers/get-program-sketch/", json={"description": description})


def get_interaction_element_id(marked_html, elem_type, description):
    """
    Gets a program sketches (full and abridged) from a description.
    :param marked_html: HTML with data-psgn-id attributes.
    :param elem_type: One of INPUT, BUTTON, or SELECT.
    :param description: A natural language description of the element.
    :return: The integer ID (data-psgn-id) of the element in the marked HTML.
    """
    assert elem_type.isupper()
    result = _api_call(
        httpx.post,
        "/transformers/get-nav-elem/",
        json={"html": marked_html, "elem_type": elem_type, "description": description},
    )["id"]
    return result


def get_schema_fields(schema):
    """
    Gets fields and their types from a given schema
    :param schema:
    :return: A dict mapping fields to their types
    """
    return _api_call(httpx.post, "/transformers/get-schema-fields/", json={"schema": schema})


def get_cleaned_data(html, schema, nodes):
    """
    Gets cleaned data from the nodes to be scraped.
    :param html: HTML of the page to scrape
    :param schema: Schema of the data to scrape
    :param nodes: Nodes to scrape
    :return: Cleaned data
    """
    return _api_call(
        httpx.post, "/transformers/get-cleaned-data/", json={"html": html, "schema": schema, "nodes": nodes}
    )


def scrape_page(html, schema, relevant_elem_ids):
    """
    Scrapes data from the provided page HTML - data will be returned in the schema provided.
    :param html: HTML of the page to scrape.
    :param schema: Schema of the data to scrape
    :param relevant_elem_ids: Ids of elements to be considered
    :return: Scraped data
    """
    return _api_call(
        httpx.post,
        "/transformers/get-custom-data/",
        json={"html": html, "schema": schema, "relevant_elem_ids": relevant_elem_ids},
    )


def get_str_about_data(data, question):
    """
    Asks GPT a question about the given data.
    :param data: the data to give GPT
    :param question: the question to ask about the data
    """
    data = _api_call(httpx.post, "/transformers/get-str-about-data/", json={"data": data, "question": question})
    return data["result"]


def get_bool_about_data(data, question):
    """
    Asks GPT a question about the given data.
    :param data: the data to give GPT
    :param question: the question to ask about the data
    """
    data = _api_call(httpx.post, "/transformers/get-bool-about-data/", json={"data": data, "question": question})
    return data["result"]


def create_pipeline(name, description, program_sketch, pseudocode, secrets):
    return _api_call(
        httpx.post,
        "/pipelines/",
        json={"name": name, "description": description, "program_sketch": program_sketch, "pseudocode": pseudocode, "secrets": secrets},
    )


def delete_pipeline(pipeline_id):
    return _api_call(httpx.delete, f"/pipelines/{pipeline_id}/")


def create_custom_function(pipeline_id, call_id, custom_function):
    _api_call(
        httpx.post,
        "/transformers/custom-function/",
        json={"pipeline": pipeline_id, "call_id": call_id, **custom_function.to_json()},
    )


def add_examples_to_custom_function(pipeline_id, call_id, custom_function, remove_old_examples):
    _api_call(
        httpx.post,
        "/transformers/custom-function/add-examples/",
        json={
            "pipeline": pipeline_id,
            "call_id": call_id,
            "remove_old_examples": remove_old_examples,
            **custom_function.to_json(),
        },
    )


def get_pipeline(pipeline_name):
    with RaiseProgramNotFound(pipeline_name):
        return _api_call(
            httpx.get,
            f"/pipelines/name/{pipeline_name}/",
        )


def get_pipelines():
    return _api_call(httpx.get, f"/pipelines/")


def get_pipeline_code(pipeline_name, variables, headless):
    with RaiseProgramNotFound(pipeline_name):
        return _api_call(
            httpx.post,
            f"/pipelines/name/{pipeline_name}/code/",
            json={
                "variables": variables,
                "headless": headless,
            },
        )


def create_pipeline_run(pipeline_id, variables, is_local):
    return _api_call(
        httpx.post,
        f"/pipelines/{pipeline_id}/runs/",
        json={"variables": variables, "is_local": is_local},
    )


def update_pipeline_run(run_id, data):
    return _api_call(
        httpx.patch,
        f"/pipelines/runs/{run_id}/",
        json=data,
    )


def get_run(run_id):
    """
    Gets details about a run
    """
    return _api_call(
        httpx.get,
        f"/pipelines/runs/{run_id}/",
    )


def poll_data(url, page_type):
    return _api_call(httpx.post, "/extract/", json={"url": url, "page_type": page_type})



import pytest

from parsagon import run
from parsagon.tests.cli_mocks import call_cli


def test_headless_remote_run_invalid(mocker, debug_logs):
    """
    Tests that we are unable to run a program in headless mode when the environment is remote, and that this is logged to the user.
    """
    call_cli(
        mocker,
        {
            "func": run,
            "program_name": "test_program",
            "headless": True,
            "remote": True,
            "verbose": False,
        },
    )
    debug_logs_lower = debug_logs.text.lower()
    assert "error" in debug_logs_lower
    assert "headless" in debug_logs_lower
    assert "remote" in debug_logs_lower



import json
import logging
import sys
from os import environ
from pathlib import Path

from parsagon.exceptions import ParsagonException

__API_BASE = environ.get("API_BASE", "https://parsagon.io").rstrip("/")
__SETTINGS_FILE = environ.get("SETTINGS_FILE", ".parsagon_profile")


logger = logging.getLogger(__name__)


def pytest_is_running():
    return "pytest" in sys.modules


def get_api_key(interactive=False):
    """
    Return API key, preventing tests from talking to the real backend
    """
    if pytest_is_running():
        return "test"

    saved_api_key = get_setting("api_key")
    if saved_api_key is not None:
        assert isinstance(saved_api_key, str), "API key must be a string."
        return saved_api_key
    elif interactive:
        while True:
            api_key = input("Please enter your Parsagon API key: ")
            if len(api_key) != 40:
                logger.error("Error: a Parsagon API key must be 40 characters long.")
            else:
                break
        save_setting("api_key", api_key)
        return api_key
    else:
        raise ParsagonException("No API key found. Please run `parsagon setup`.")


def save_settings(settings):
    with open(get_settings_file_path(), "w") as f:
        json.dump(settings, f, indent=4, sort_keys=True)


def save_setting(name, value):
    settings = get_settings()
    settings[name] = value
    save_settings(settings)


def get_settings():
    if get_settings_file_path().is_file():
        with open(get_settings_file_path(), "r") as f:
            return json.load(f)
    else:
        return {}


def clear_settings():
    if get_settings_file_path().is_file():
        get_settings_file_path().unlink()


def get_setting(key):
    return get_settings().get(key)


def get_settings_file_path():
    """
    Return settings file path, which is a hidden file in the user's home directory
    """
    return Path().home() / __SETTINGS_FILE


def get_api_base():
    """
    Return API base, preventing tests from talking to the real backend
    """
    if pytest_is_running():
        return "http://test"
    else:
        return __API_BASE


def get_logging_config(log_level="INFO"):
    return {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "standard": {"format": "%(message)s"},
        },
        "handlers": {
            "default": {
                "level": "DEBUG",
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "": {
                "handlers": ["default"],
                "level": log_level,
                "propagate": False,
            },  # root logger
            "parsagon": {
                "handlers": ["default"],
                "level": log_level,
                "propagate": False,
            },
        },
    }



from typing import Dict, List, Any


class CustomFunction:
    """
    A custom function to be converted to one or more transformers on the backend
    """

    def __init__(self, name: str, arguments: Dict[str, Any], examples: List[Dict]):
        """
        :param name: The name of the custom function as it is known to GPT.
        :param arguments: Arguments used for the actual performing of the function.
        :param examples: Holds other values useful for the creation of transformers.
        """
        self.name = name
        self.arguments = arguments
        self.examples = examples

    def to_json(self):
        return {
            "name": self.name,
            "arguments": self.arguments,
            "examples_data": self.examples,
        }



import pytest

from parsagon.secrets import extract_secrets


def test_non_secrets_are_not_extracted():
    """
    Non-secrets should not be extracted from task descriptions.
    """
    task = 'Go to https://example.com. Type {username: "myusername"} in the username field'
    task, secrets = extract_secrets(task)
    assert secrets == {}
    assert task == 'Go to https://example.com. Type {username: "myusername"} in the username field'


def test_secret_is_extracted():
    """
    A secret should be extracted and replaced in a task description.
    """
    task = 'Go to https://example.com. Type {SECRET_PASSWORD: "mypassword"} in the password field'
    task, secrets = extract_secrets(task)
    assert secrets == {"SECRET_PASSWORD": "mypassword"}
    assert task == 'Go to https://example.com. Type {SECRET_PASSWORD: "******"} in the password field'


def test_secret_with_quotes_is_extracted():
    """
    A secret with quotes in its value should be extracted and replaced in a task description.
    """
    task = 'Go to https://example.com. Type {SECRET_PASSWORD: "mypassword\\"?!1"} in the password field'
    task, secrets = extract_secrets(task)
    assert secrets == {"SECRET_PASSWORD": 'mypassword"?!1'}
    assert task == 'Go to https://example.com. Type {SECRET_PASSWORD: "******"} in the password field'


def test_multiple_secrets_are_extracted():
    """
    Multiple secrets should be extracted and replaced in the same task description.
    """
    task = 'Go to https://example.com. Type {SECRET_PASSWORD: "mypassword"} in the password field. Type {SECRET_ADDRESS: "myaddress"} in the address field'
    task, secrets = extract_secrets(task)
    assert secrets == {"SECRET_PASSWORD": "mypassword", "SECRET_ADDRESS": "myaddress"}
    assert task == 'Go to https://example.com. Type {SECRET_PASSWORD: "******"} in the password field. Type {SECRET_ADDRESS: "******"} in the address field'


def test_secrets_mixed_with_non_secrets_are_extracted():
    """
    Multiple secrets should be extracted and replaced in the same task description, and non-secrets should remain the same.
    """
    task = 'Go to https://example.com. Type {USERNAME: "myusername"} in the username field. Type {SECRET_PASSWORD: "mypassword"} in the password field. Type {SECRET_ADDRESS: "myaddress"} in the address field'
    task, secrets = extract_secrets(task)
    assert secrets == {"SECRET_PASSWORD": "mypassword", "SECRET_ADDRESS": "myaddress"}
    assert task == 'Go to https://example.com. Type {USERNAME: "myusername"} in the username field. Type {SECRET_PASSWORD: "******"} in the password field. Type {SECRET_ADDRESS: "******"} in the address field'
